/****************************************************************************
 *  Copyright 2003-2010 Dorian C. Arnold, Philip C. Roth, Barton P. Miller  *
 *                  Detailed MRNet usage rights in "LICENSE" file.          *
 ****************************************************************************/

#include "utils.h"
#include "mrnet/Event.h"
#include "xplat/SocketUtils.h"

using namespace std;

namespace MRN {

// generic write-to-notify-pipe callback
void PipeNotifyCallbackFn( Event*, void* idata )
{
    EventPipe* ep = (EventPipe*)idata;
    ep->signal();
}

EventPipe::EventPipe()
    : _has_data(false)
{
    set_ReadFd( -1 );
    set_WriteFd( -1 );
#if !defined(os_windows)
    int pipeFDs[2];
    int ret = pipe(pipeFDs);
    if( ret == 0 ) {
        _sync.Lock();
        set_ReadFd( pipeFDs[0] );
        set_WriteFd( pipeFDs[1] );
        _sync.Unlock();
    }
    else
        mrn_printf(FLF, stderr, "pipe() failed - %s", strerror(errno));
#endif
}

EventPipe::~EventPipe()
{
#if !defined(os_windows)
    _sync.Lock();
    if( _pipe_fds[0] != -1 ) {
        XPlat::SocketUtils::Close( _pipe_fds[0] );
        set_ReadFd( -1 );
    }
    if( _pipe_fds[1] != -1 ) {
        XPlat::SocketUtils::Close( _pipe_fds[1] );
        set_WriteFd( -1 );
    }
    _sync.Unlock();
#endif
}

void EventPipe::signal(void)
{
#if !defined(os_windows)
    _sync.Lock();
    if(( get_WriteFd() == -1 ) || _has_data ) {
        _sync.Unlock();
        return;
    }
    _sync.Unlock();

    mrn_dbg(5, mrn_printf(FLF, stderr, "writing pipefd\n" ));
    char c = '!';
    int ret = write( get_WriteFd(), &c, sizeof(char) );
    if( ret == -1 )
        mrn_printf(FLF, stderr, "write() failed - %s", strerror(errno));
    else {
        _sync.Lock();
        _has_data = true;;
        _sync.Unlock();
    }
#endif
}

void EventPipe::clear(void)
{
#if !defined(os_windows)
    _sync.Lock();
    if(( get_ReadFd() == -1 ) || ( ! _has_data )) {
        _sync.Unlock();
        return;
    }
    _sync.Unlock();

    mrn_dbg(5, mrn_printf(FLF, stderr, "clearing pipefd\n" ));
    char c;
    int ret = read( get_ReadFd(), &c, sizeof(char) );
    if( ret == -1 )
        mrn_printf(FLF, stderr, "read() failed - %s", strerror(errno));
    else {
        _sync.Lock();
        _has_data = false;
        _sync.Unlock();
    }
#endif
}

// static Event members
EventClass Event::EVENT_CLASS_ALL = -1;
EventClass Event::DATA_EVENT      = 0;
EventClass Event::TOPOLOGY_EVENT  = 1;
EventClass Event::ERROR_EVENT     = 2;

EventType Event::EVENT_TYPE_ALL = -1;

EventType DataEvent::DATA_AVAILABLE = 0;

EventType ErrorEvent::ERROR_INTERNAL = 0;
EventType ErrorEvent::ERROR_SYSTEM   = 1;
EventType ErrorEvent::ERROR_USAGE    = 2;

EventType TopologyEvent::TOPOL_ADD_BE        = 0;
EventType TopologyEvent::TOPOL_ADD_CP        = 1;
EventType TopologyEvent::TOPOL_REMOVE_NODE   = 2;
EventType TopologyEvent::TOPOL_CHANGE_PARENT = 3;

EventMgr::EventMgr()
{
    reset_Callbacks();
}

EventMgr::~EventMgr()
{
    clear_Events();
    _cbs.clear();
}

void EventMgr::reset_Callbacks()
{
    // initialize the event class/type callback maps
    EventType min, max;
    evt_cb_list empty_list;
    int u;

    min = DataEvent::min_Type();
    max = DataEvent::max_Type();
    evt_typ_cb_map& data_map = _cbs[ Event::DATA_EVENT ];
    for( u = min; u <= max; u++ )
        data_map[ u ] = empty_list;

    min = ErrorEvent::min_Type();
    max = ErrorEvent::max_Type();
    evt_typ_cb_map& error_map = _cbs[ Event::ERROR_EVENT ];
    for( u = min; u <= max; u++ )
        error_map[ u ] = empty_list;

    min = TopologyEvent::min_Type();
    max = TopologyEvent::max_Type();
    evt_typ_cb_map& topol_map = _cbs[ Event::TOPOLOGY_EVENT ];
    for( u = min; u <= max; u++ )
        topol_map[ u ] = empty_list;
}

bool EventMgr::have_Events() const
{
    return ! _evts.empty();
}

unsigned int EventMgr::get_NumEvents() const
{
    return _evts.size();
}
    
bool EventMgr::add_Event( Event* ievt )
{
    if( ievt != NULL ) {
        _evts.push_back( ievt );
        execute_Callbacks( ievt );
        return true;
    }
    return false;
}

Event* EventMgr::get_NextEvent()
{
    if( _evts.size() ) {
        Event* ret = _evts.front();
        _evts.pop_front();
        return ret;
    }
    else
        return NULL;
}

void EventMgr::clear_Events()
{
    _evts.clear();
}

bool EventMgr::register_Callback( EventClass iclass, EventType ityp,
                                  evt_cb_func ifunc, void* idata )
{
    if( ifunc == (evt_cb_func)NULL )
        return false;

    evt_cb_info cb_info = make_pair( ifunc, idata );

    std::map< EventClass, evt_typ_cb_map >::iterator i = _cbs.find( iclass );
    if( i == _cbs.end() )
        return false;
 
    evt_typ_cb_map& ins_map = i->second;
    evt_typ_cb_map::iterator mapi;
    if( ityp == Event::EVENT_TYPE_ALL ) {
        mapi = ins_map.begin();
        for( ; mapi != ins_map.end(); mapi++ ) {
            evt_cb_list& elist = mapi->second;
            elist.push_back( cb_info );
        }
    }
    else {
        mapi = ins_map.find(ityp);                
        if( mapi == ins_map.end() )
            return false;

        evt_cb_list& ins_list = mapi->second;
        ins_list.push_back( cb_info );
    }
    return true;
}
  
bool EventMgr::remove_Callbacks( EventClass iclass, EventType ityp )
{
    if( iclass == Event::EVENT_CLASS_ALL ) {
        if( ityp == Event::EVENT_TYPE_ALL ) {
            reset_Callbacks();
            return true;
        }
        else {
            // invalid use
            return false;
        }
    }

    return remove_Callback( (evt_cb_func)NULL, iclass, ityp );
}

bool EventMgr::remove_Callback( evt_cb_func ifunc, EventClass iclass, EventType ityp )
{
    std::map< EventClass, evt_typ_cb_map >::iterator i;
    i = _cbs.find( iclass );
    if( i == _cbs.end() )
        return false;

    evt_typ_cb_map& rm_map = i->second;
    evt_typ_cb_map::iterator mapi;
    if( ityp == Event::EVENT_TYPE_ALL ) {
        //Remove for all event type
        if( ifunc == (evt_cb_func)NULL ) {
            //Remove all functions
            for( mapi = rm_map.begin(); mapi != rm_map.end(); mapi++ )
                mapi->second.clear();
            return true;
        }
        else {
            //Remove only ifunc 
            bool rm_flag = false;
            for( mapi = rm_map.begin(); mapi != rm_map.end(); mapi++ ) {
                bool retval = remove_CallbackFunc( ifunc, mapi->second );
                if( retval )
                    rm_flag = true;
            }
            return rm_flag;
        }
    }
    else {
        //Remove for only ityp event type
        mapi = rm_map.find(ityp);
        if( mapi == rm_map.end() )
            return false;
         
        if( ifunc == (evt_cb_func)NULL ) {
            //Remove all functions
            mapi->second.clear();
            return true;
        }
        else {
            //Remove only ifunc
            return remove_CallbackFunc( ifunc, mapi->second );
        }
    }
    // shouldn't get here
    return false;
}
 
bool EventMgr::remove_CallbackFunc( evt_cb_func ifunc,
                                    evt_cb_list& ifuncs )
{
    evt_cb_list::iterator rm_li = ifuncs.begin();
    for( ; rm_li != ifuncs.end(); rm_li++ ) {
        if( rm_li->first == ifunc ) {
            ifuncs.erase( rm_li );
            return true;
        }
    }
    return false;
}
 
bool EventMgr::execute_Callbacks( Event* ievt )
{
    EventClass c = ievt->get_Class();
    EventType t = ievt->get_Type();

    map< EventClass, evt_typ_cb_map >::iterator i;
    i = _cbs.find( c );
    if( i == _cbs.end() )
        return false;
    else {
        evt_typ_cb_map& etyp_map = i->second;
        evt_typ_cb_map::iterator mapi = etyp_map.find( t );
        if( mapi == etyp_map.end() )
            return false;
        else {
            evt_cb_list& elist = mapi->second;
            evt_cb_list::iterator li;
            for( li = elist.begin(); li != elist.end(); li++ ) {
                evt_cb_func f = li->first;
                void* f_data = li->second;
                (*f)( ievt, f_data );
            }
        }
    }
 
    return true;
}


} // namespace MRN
